@@ -1,6 +1,4 @@
-# -*- coding: utf-8 -*-
-#
-# Copyright 2010-2019 The pygit2 contributors
+# Copyright 2010-2020 The pygit2 contributors
 #
 # This file is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2,
@@ -25,14 +23,11 @@
 # the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 # Boston, MA 02110-1301, USA.
 
-
-# Import from the future
-from __future__ import absolute_import, unicode_literals
-
-from _pygit2 import Oid
+from ._pygit2 import Oid
 from .errors import check_error
 from .ffi import ffi, C
 
+
 class Submodule(object):
 
     @classmethod
@@ -84,4 +79,3 @@ class Submodule(object):
         """Head of the submodule."""
         head = C.git_submodule_head_id(self._subm)
         return Oid(raw=bytes(ffi.buffer(head)[:]))
-
